#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include "sysstuff.h"
#include "typedefs.h"
#include "macros.h"
#include "smalloc.h"
#include "assert.h"
#include "physics.h"
#include "macros.h"
#include "vec.h"
#include "force.h"
#include "invblock.h"
#include "confio.h"
#include "nsb.h"
#include "names.h"
#include "network.h"
#include "wnblist.h"
#include "pbc.h"
#include "ns.h"
#include "nrnb.h"
#include "bondf.h"
#include "mshift.h"
#include "txtdump.h"
#include "copyrite.h"
#include <stdio.h>
#include <string.h>
#include "fatal.h"
#include "typedefs.h"
#include "gmx_random.h"
#include "lambdadynamics.h"
#include <stdlib.h>
#ifdef GMX_MPI
#include <mpi.h>
#endif

int ts_array = 4;
double adaption_treshold = 15.0; /* in picoseconds */
double group_count = 0; /* number of lambda groups */


/* ftegele: gauss distributed random number generator
   based on the box mueller method

   ASSUMES AN INITIALIZED RANDOM NUMBER GENERATOR
   for compatbility reasons function rand() is used */

/* this version is the polar form   */
/* similar to the orginal gaussdist */
real rnd_gauss_box_mueller(void) {
    real x1, x2, w, y1, y2; 
         do {
           x1 = 2.0 * rand()/RAND_MAX - 1.0;
           x2 = 2.0 * rand()/RAND_MAX - 1.0;
            w = x1 * x1 + x2 * x2;
         } while ( w >= 1.0 );

         w = sqrt( (-2.0 * log( w ) ) / w );
         //y2 = x2 * w;
    return x1*w;
}


real gaussdist(gmx_rng_t rng,real sigma){
  real r=2.0,x,y;
  do{
    x = 2.0*(real)gmx_rng_uniform_real(rng)-1.0;
    y = 2.0*(real)gmx_rng_uniform_real(rng)-1.0;
    r = x*x+y*y;
  }while(r>1.0||r==0.0);
  r = x*sqrt(-2.0*log(r)/r);
  r = r*sigma;
  return (r);
}

// returns -1 on error, 1 on EOF, 0 else
int read_ph_data(FILE* dat, struct ph_adapt* ph) {
    char temp[255];
    double dum=0.0;

    ph->next = (struct ph_adapt*)  malloc (sizeof (struct ph_adapt));
    ph->residue = malloc( sizeof(char) * 129); // alloc 129 byte to store 128 byte long residue names...

    /* temp could be checked whether correct information is entered */
    if (fscanf(dat, "residue = %s;\n",ph->residue) == EOF)
	return 1;
    /* Always read in as double, then convert to real */
    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->barrier = dum;

    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->const_a = dum;

    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->ph_a = dum;

    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->const_2 = dum;

    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->const_3 = dum;

    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->const_b = dum;

    if (fscanf(dat, "%s = %lg;\n", temp, &dum) == EOF)
        return -1;
    else
        ph->ph_b = dum;

//    fprintf(stderr, "Read in: %s, %g, %g, %g, %g, %g, %g\n", ph->residue, ph->const_a, ph->ph_a, ph->const_2,ph->const_3,ph->const_b, ph->ph_b);
    if ( (strncmp( (ph->residue),"HIS",3)==0) ) 
      ph->second_pka=14.1;
    else
      ph->second_pka=0;
	
    fprintf(stderr, "Read in:second_pka=%g\n",ph->second_pka);

    return 0;
}

/* allocates the memory and initializes data for the multi lambda control */
/* assumes that groups provided in the data file are correctly mapping
   to atoms perturbed in the topology */
void init_mlrec(t_mdatoms *atoms, t_commrec *cr, t_inputrec *ir){
  int xpos,i;
  FILE* dat;
  char temp[255];
  char outfilename[255];
  char top2file[255];
  struct multi_lambda *tmp = NULL;
  struct multi_lambda *old = NULL;
  struct multi_lambda *tmp_temp = NULL;
  int cont_reading = 1;
  real initial_lambda = 0.0;
  FILE* dat_his;
  double micro_pka1, micro_pka2;

  if((dat = fopen("lambda_groups.dat","r")) == NULL)
        fprintf(stderr, "ERROR OPENING lambda_groups.dat!!!!\n");

  /* it is assumed that this file is not empty */
  /* structure of the file:
      name = %s
      initial lambda = %lg
      number of atoms = %i
      atom1 atom2 atom3 atom4 ...
      
      name = %s
      .
      .
      .
   */

//  fprintf(stderr, "Version 0.95 - adaptive Barrier potential and all the cool stuff\n");
  fprintf(stderr, "Version 1.3 - Constant pH MD with Gromacs 3.3.1\n");
  fprintf(stderr, "atoms->nPerturbed %d\n",atoms->nPerturbed);

  //if (atoms->nPerturbed > 0) //there are perturbed atoms
  //{
    ml_rec = tmp = malloc ( sizeof (struct multi_lambda));

    while (cont_reading == 1)
    { 
      fprintf(stderr, "Reading in with cont_reading = %i\n", cont_reading);
      
      tmp->name = malloc ( sizeof ( char) * 256 );
      tmp->lambda = malloc ( sizeof ( struct t_ldrec));
      
      /* use the index file to read in the different lambda dynamics groups */
      /* assume we have read in atoms 3,4,5,6 */

      if (fscanf(dat, "%s = %s\n",temp, tmp->name) != 2)
      {
        /* if here is an readin failure, this means the file is at its end,
           simply exit */
 fprintf(stderr,"Reading is ending here... %s und %s\n", temp, tmp->name);
        cont_reading = 0;
      }
      else
      {
        /* there is no error */
	//sprintf(outfilename,""); //define the output file name for this lambda group 
        //tmp->out = ffopen("lambdadynamics.xvg","w");  

        fscanf(dat, "%s = %i\n", temp, &(tmp->residue_number));
		fprintf(stderr,"%s\n", temp);
        fscanf(dat, "%s = %g\n", temp, &initial_lambda);
		fprintf(stderr,"%s\n", temp);
        fscanf(dat, "%s = %i\n", temp, &(tmp->nr_atoms));
		fprintf(stderr,"%s\n", temp);
        fprintf(stderr, "Reading in %i atoms for lambda group %s with initial lambda = %g\n", tmp->nr_atoms, tmp->name, initial_lambda);

        group_count+=1.0;

        tmp->atoms = malloc ( sizeof(int) * tmp->nr_atoms);
        tmp->chargeA = malloc ( sizeof(real) * tmp->nr_atoms);
        tmp->chargeB = malloc ( sizeof(real) * tmp->nr_atoms);
        tmp->massA = malloc ( sizeof(real) * tmp->nr_atoms);
        tmp->massB = malloc ( sizeof(real) * tmp->nr_atoms);
        tmp->typeA = malloc ( sizeof (int) * tmp->nr_atoms);
        tmp->typeB = malloc ( sizeof (int) * tmp->nr_atoms);

        for (xpos=0; xpos<tmp->nr_atoms; xpos++)
        {
          if (fscanf(dat, "%i", &(tmp->atoms[xpos])) == EOF)
            fprintf(stderr,"FATAL ERROR: Crap in the input file for lambda groups\n");
	  else
	  {
	     /* readin successful, check whether this atom is really perturbed */
            tmp->atoms[xpos] = tmp->atoms[xpos]-1; /* counting is shifted by one */
            fprintf(stderr,"atom=%i, perturbed=%i\n", tmp->atoms[xpos], atoms->bPerturbed[tmp->atoms[xpos]]);
	  }
	 
          tmp->chargeA[xpos] = atoms->chargeA[tmp->atoms[xpos]]; //initial charge
          tmp->chargeB[xpos] = atoms->chargeB[tmp->atoms[xpos]]; //initial charge
          tmp->massA[xpos] = atoms->massA[tmp->atoms[xpos]]; //initial mass
          tmp->massB[xpos] = atoms->massB[tmp->atoms[xpos]]; //initial mass
          tmp->typeA[xpos] = atoms->typeA[tmp->atoms[xpos]]; //initial typeA
          tmp->typeB[xpos] = atoms->typeB[tmp->atoms[xpos]]; //initial typeB
	  fprintf(stderr, "Charges: %lg to %lg, Masses: %lg to %lg Types: %d to %d\n",  tmp->chargeA[xpos],  tmp->chargeB[xpos],  tmp->massA[xpos],  tmp->massB[xpos],tmp->typeA[xpos],tmp->typeB[xpos]);
        }

	/* Overwrite the charges (and masses, but not types) of HISB according to file HIS2.top */
	/* The topologies (charges) of the four possible HIS are now stored in chargeA (HISH) and chargeB (HISA) of residue HISA */
	/* (read from topol.top) and in chargeA (HISB) and chargeB (HISnegative) of HISB (read from HIS2.top). */

	if ( (strncmp( (tmp->name),"HISB",4)==0) ) /* 4 states model */
        {
	    if((dat_his = fopen("HISB.top","r")) == NULL)
		    fprintf(stderr, "ERROR OPENING HISB.top!!!!\n");
	    fscanf(dat_his, "%s %s %s %s %s %s %s\n",temp,temp,temp,temp,temp,temp,temp);
	    for (xpos=0; xpos<tmp->nr_atoms; xpos++)
	    {
		fscanf(dat_his, "%i %s %g %g %s %g %g\n",&i,temp,&(tmp->chargeA[xpos]),&(tmp->massA[xpos]),temp,&(tmp->chargeB[xpos]),&(tmp->massB[xpos]) );
		fprintf(stderr,"HISB atom %d: %g %g %g %g\n",xpos,tmp->chargeA[xpos],tmp->massA[xpos],tmp->chargeB[xpos],tmp->massB[xpos]);
	    } 
	    fclose(dat_his);
	}
	if (  (strncmp( (tmp->name),"GL2B",4)==0) || (strncmp( (tmp->name),"AS2B",4)==0) || (strncmp( (tmp->name),"HI2B",4)==0) ) /* 3 states model; same stuff as 4 states */
        {
	    sprintf(top2file,"%s.top",tmp->name);
	    if((dat_his = fopen(top2file,"r")) == NULL)
	      fprintf(stderr, "ERROR OPENING %s!!!!\n",top2file);
	    fscanf(dat_his, "%s %s %s %s %s %s %s\n",temp,temp,temp,temp,temp,temp,temp);
	    for (xpos=0; xpos<tmp->nr_atoms; xpos++)
	    {
		fscanf(dat_his, "%i %s %g %g %s %g %g\n",&i,temp,&(tmp->chargeA[xpos]),&(tmp->massA[xpos]),temp,&(tmp->chargeB[xpos]),&(tmp->massB[xpos]) );
		fprintf(stderr,"%s atom %d: %g %g %g %g\n",tmp->name,xpos,tmp->chargeA[xpos],tmp->massA[xpos],tmp->chargeB[xpos],tmp->massB[xpos]);
	    } 
	    fclose(dat_his);
	}
	
	/* we want to overwrite types, bonds and angles for GL2 and AS2 (only A state, B is uncharged, therefore same as B state in topol.top) */
	//	if (  (strncmp( (tmp->name),"GL2B",4)==0) || (strncmp( (tmp->name),"AS2B",4)==0) )
        //{
	/* this is for the types */
	//for (xpos=0; xpos<tmp->nr_atoms; xpos++) /* for gromos topologies: oxygens at xpos 1 and 3 */
	/*    {
	      if (xpos==1){
	        tmp->typeA[xpos]=atoms->typeA[tmp->atoms[xpos+2]];
                fprintf (stderr,"%s atom %d: typeA from %d to %d\n",tmp->name,xpos,atoms->typeA[tmp->atoms[xpos]],tmp->typeA[xpos]);
	      }	
	      if (xpos==3) {
	        tmp->typeA[xpos]=atoms->typeA[tmp->atoms[xpos-2]];
                fprintf (stderr,"%s atom %d: typeA from %d to %d\n",tmp->name,xpos,atoms->typeA[tmp->atoms[xpos]],tmp->typeA[xpos]);
	      }	
	    } 
	}*/
	

        /* initalize every ldrec on its own (i know this is not perfect ;) */
        ir->init_lambda = initial_lambda;
        init_ldrec(cr, ir, tmp->lambda, tmp->name); //then the ph dependency data could be saved directly there?


        /* set the initial dvdl to zero */
        ml_rec->dvdl = 0;
        
        /* set tmp to point on its next element */
	tmp->next = malloc ( sizeof (struct multi_lambda));
        tmp->his_partner=NULL;
        old = tmp;
        tmp = tmp->next;
      }
    }
    //}
  old->next = NULL;

  tmp=ml_rec;
      while (tmp != NULL) {
	  tmp_temp=tmp;
	  if (strncmp(tmp->name,"HISA",4)==0) { /* 4 states model*/
	      while (tmp_temp != NULL) {
	         if( ( (strncmp( (tmp_temp->name),"HISB",4)==0) ) && ((tmp_temp->residue_number)==(tmp->residue_number)) ) { 
		   fprintf(stderr, "In IF:%s res num=%i and %s res num=%i\n",tmp->name,tmp->residue_number, tmp_temp->name,tmp_temp->residue_number);
         	    tmp->his_partner=tmp_temp;
		    tmp->lambda->his_partner_const_b=tmp->his_partner->lambda->phc->const_b;
		    tmp_temp->his_partner=tmp;
		    tmp_temp->lambda->his_partner_const_b=tmp->lambda->phc->const_b;
		    fprintf(stderr, "his_partner_const_b= %g, partner's his_partner_const_b= %g \n",tmp->lambda->his_partner_const_b,tmp_temp->lambda->his_partner_const_b);
                    tmp_temp = NULL;
	         }
		 else
                    tmp_temp=tmp_temp->next; 
              }
          }
          old = tmp;
          tmp = tmp->next;   
      }
      old->next = NULL;

  tmp=ml_rec;
      while (tmp != NULL) {
	  tmp_temp=tmp;
	  if (strncmp(tmp->name,"GL2A",4)==0) {  /* 3 states model; same stuff as 4 states plus micro versus macro issue */
	    while (tmp_temp != NULL) {
	         if( ( (strncmp( (tmp_temp->name),"GL2B",4)==0) ) && ((tmp_temp->residue_number)==(tmp->residue_number)) ) { 
		   fprintf(stderr, "sono in IF:%s res num=%i and %s res num=%i\n",tmp->name,tmp->residue_number, tmp_temp->name,tmp_temp->residue_number);
         	    tmp->his_partner=tmp_temp;
		    tmp->lambda->his_partner_const_b=tmp->his_partner->lambda->phc->const_b;
		    tmp_temp->his_partner=tmp;
		    tmp_temp->lambda->his_partner_const_b=tmp->lambda->phc->const_b;
		    micro_pka1 = (double) -(tmp->lambda->phc->const_b) / (tmp->lambda->phc->ph_b); 
		    micro_pka2 = (double) -(tmp_temp->lambda->phc->const_b) / (tmp_temp->lambda->phc->ph_b); 
		    tmp->lambda->micro_diff= - (tmp->lambda->phc->ph_b)*(micro_pka2 - micro_pka1);
		    tmp_temp->lambda->micro_diff = tmp->lambda->micro_diff;
		    tmp->lambda->macro_pka = log10((exp10(micro_pka1)+exp10(micro_pka2)));
		    tmp_temp->lambda->macro_pka = tmp->lambda->macro_pka;
		    fprintf(stderr, "his_partner_const_b= %g, partner's his_partner_const_b= %g \n",tmp->lambda->his_partner_const_b,tmp_temp->lambda->his_partner_const_b);
		    fprintf(stderr, "GL2A/GL2B residue %d pka_macro= %lg %lg\n", tmp->residue_number, tmp->lambda->macro_pka,tmp_temp->lambda->macro_pka );
		    fprintf(stderr, "pka_micro_GL2B - pka_micro_GL2A= %lg - %lg = %lg %lg\n",(-((tmp->lambda->phc->ph_b))*(micro_pka2)),(-((tmp->lambda->phc->ph_b))*(micro_pka1)),tmp->lambda->micro_diff,tmp_temp->lambda->micro_diff);
                    tmp_temp = NULL;
	         }
		 else
                    tmp_temp=tmp_temp->next; 
              }
          }
          old = tmp;
          tmp = tmp->next;   
      }
      old->next = NULL;

  tmp=ml_rec;
      while (tmp != NULL) {
	  tmp_temp=tmp;
	  if (strncmp(tmp->name,"AS2A",4)==0) {  /* 3 states model; same stuff as 4 states plus micro versus macro issue */
	    while (tmp_temp != NULL) {
	         if( ( (strncmp( (tmp_temp->name),"AS2B",4)==0) ) && ((tmp_temp->residue_number)==(tmp->residue_number)) ) { 
		   fprintf(stderr, "sono in IF:%s res num=%i and %s res num=%i\n",tmp->name,tmp->residue_number, tmp_temp->name,tmp_temp->residue_number);
         	    tmp->his_partner=tmp_temp;
		    tmp->lambda->his_partner_const_b=tmp->his_partner->lambda->phc->const_b;
		    tmp_temp->his_partner=tmp;
		    tmp_temp->lambda->his_partner_const_b=tmp->lambda->phc->const_b;
		    micro_pka1 = (double) -(tmp->lambda->phc->const_b) / (tmp->lambda->phc->ph_b); 
		    micro_pka2 = (double) -(tmp_temp->lambda->phc->const_b) / (tmp_temp->lambda->phc->ph_b); 
		    tmp->lambda->micro_diff= - (tmp->lambda->phc->ph_b)*(micro_pka2 - micro_pka1);
		    tmp_temp->lambda->micro_diff = tmp->lambda->micro_diff;
		    tmp->lambda->macro_pka = log10((exp10(micro_pka1)+exp10(micro_pka2)));
		    tmp_temp->lambda->macro_pka = tmp->lambda->macro_pka;
		    fprintf(stderr, "his_partner_const_b= %g, partner's his_partner_const_b= %g \n",tmp->lambda->his_partner_const_b,tmp_temp->lambda->his_partner_const_b);
		    fprintf(stderr, "AS2A/AS2B residue %d pka_macro= %lg %lg\n", tmp->residue_number, tmp->lambda->macro_pka,tmp_temp->lambda->macro_pka );
		    fprintf(stderr, "pka_micro_AS2B - pka_micro_AS2A= %lg - %lg = %lg %lg\n",(-((tmp->lambda->phc->ph_b))*(micro_pka2)),(-((tmp->lambda->phc->ph_b))*(micro_pka1)),tmp->lambda->micro_diff,tmp_temp->lambda->micro_diff);
                    tmp_temp = NULL;
	         }
		 else
                    tmp_temp=tmp_temp->next; 
              }
          }
          old = tmp;
          tmp = tmp->next;   
      }
      old->next = NULL;

  tmp=ml_rec;
      while (tmp != NULL) {
	  tmp_temp=tmp;
	  if (strncmp(tmp->name,"HI2A",4)==0) {  /* 3 states model; same stuff as 4 states */
	    while (tmp_temp != NULL) {
	         if( ( (strncmp( (tmp_temp->name),"HI2B",4)==0) ) && ((tmp_temp->residue_number)==(tmp->residue_number)) ) { 
		   fprintf(stderr, "sono in IF:%s res num=%i and %s res num=%i\n",tmp->name,tmp->residue_number, tmp_temp->name,tmp_temp->residue_number);
         	    tmp->his_partner=tmp_temp;
		    tmp->lambda->his_partner_const_b=tmp->his_partner->lambda->phc->const_b;
		    tmp_temp->his_partner=tmp;
		    tmp_temp->lambda->his_partner_const_b=tmp->lambda->phc->const_b;
		    micro_pka1 = (double) -(tmp->lambda->phc->const_b) / (tmp->lambda->phc->ph_b); 
		    micro_pka2 = (double) -(tmp_temp->lambda->phc->const_b) / (tmp_temp->lambda->phc->ph_b); 
		    tmp->lambda->micro_diff= - (tmp->lambda->phc->ph_b)*(micro_pka2 - micro_pka1);
		    tmp_temp->lambda->micro_diff = tmp->lambda->micro_diff;
		    tmp->lambda->macro_pka = -(log10((exp10(-micro_pka1)+exp10(-micro_pka2))));
		    tmp_temp->lambda->macro_pka = tmp->lambda->macro_pka;
		    fprintf(stderr, "his_partner_const_b= %g, partner's his_partner_const_b= %g \n",tmp->lambda->his_partner_const_b,tmp_temp->lambda->his_partner_const_b);
		    fprintf(stderr, "HI2A/HI2B residue %d pka_macro= %lg %lg\n", tmp->residue_number, tmp->lambda->macro_pka,tmp_temp->lambda->macro_pka );
		    fprintf(stderr, "pka_micro_HI2B - pka_micro_HI2A= %lg - %lg = %lg %lg\n",(-((tmp->lambda->phc->ph_b))*(micro_pka2)),(-((tmp->lambda->phc->ph_b))*(micro_pka1)),tmp->lambda->micro_diff,tmp_temp->lambda->micro_diff);
                    tmp_temp = NULL;
	         }
		 else
                    tmp_temp=tmp_temp->next; 
              }
          }
          old = tmp;
          tmp = tmp->next;   
      }
      old->next = NULL;
      //}//if perturbed atoms
  fclose(dat);
  
}


void init_ldrec(t_commrec *cr, t_inputrec *ir, struct t_ldrec *ld, char* name){
  
    FILE* dat;
    FILE* temp_ph;
    FILE* param_his;
    int res = 0;
    struct ph_adapt* curr;
    int i,j;
    char param2file[255];
    char fullnameres[255];
    i=0;j=0;
    double dum=0.0;
    real radius = 0.55;

  ld->x0 = ld->x = ir->init_lambda;
  ld->v = 0;
  ld->T_ref = ir->T_lambda; /* new variable */
  ld->T = 0.0; /* instantaneous temp at t=0 */
  ld->m = ir->m_lambda; /* new variable */
  ld->ekin = 0;
  ld->nu_T = (ir->delta_t)*(ir->nu_T_lambda); /* a value of nu_T_lambda = 5 means: Every fifth timestep*/

  /* theta from lambda_init */
  ld->thetha = acos((ir->init_lambda-0.5)/radius);
//  ld->thetha = acos(2*(ir->init_lambda)-1);

  ld->v_thetha = 0; /* no initial kinetic energy */

  /* init ph_depend */
  ld->ph_value = 5.5;
  ld->phc = (struct ph_adapt*)  malloc (sizeof (struct ph_adapt));

  ld->transitions = (struct lambda_ts*) malloc (sizeof (struct lambda_ts) * ts_array);

  /* get this data from the file pka_data.dat */
  if((dat = fopen("pka_data.dat","r")) == NULL)
      fprintf(stderr, "ERROR OPENING pka_data.dat!!!!\n");
  curr = ld->phc;
  while ((res=read_ph_data(dat, curr)) == 0)
      curr = curr->next;
  
  switch (res) {
    case -1:
        break;
    case 1:
	break;
    case 0:
	free(curr->next);
	curr->next = NULL;
  }

  fclose(dat);

  curr=ld->phc;
  while (curr != NULL) {
      fprintf(stderr,"curr->residue=%s,name in pka_data.dat=%s\n",curr->residue,name);
      if (strncmp(curr->residue,name,4)==0) {
	  ld->phc=curr;
       	  fprintf(stderr, "Read in: %s, %g, %g, %g, %g, %g, %g, %g\n",ld->phc->residue,ld->phc->barrier,ld->phc->const_a,ld->phc->ph_a,ld->phc->const_2,ld->phc->const_3,ld->phc->const_b,ld->phc->ph_b);
	  if ( (strncmp(curr->residue,"HISA",4)==0) || (strncmp(curr->residue,"HISB",4)==0) || (strncmp(curr->residue,"GL2A",4)==0) || (strncmp(curr->residue,"GL2B",4)==0) || (strncmp(curr->residue,"AS2A",4)==0) || (strncmp(curr->residue,"AS2B",4)==0) || (strncmp(curr->residue,"HI2A",4)==0) || (strncmp(curr->residue,"HI2B",4)==0)  ) { /* 4 and 3 states model */
	      strncpy(fullnameres,curr->residue,4);
	      fullnameres[4] = '\0';
	      sprintf(param2file,"param_%s.dat",fullnameres);
	      fprintf(stderr, "param2file=%s,fullnameres=%s\n",param2file,fullnameres);
	      if((param_his = fopen(param2file,"r")) == NULL)
		fprintf(stderr, "ERROR OPENING %s!!!!\n",param2file);
	      fprintf(stderr, "Read in: %s,",ld->phc->residue);
	      for (i=0;i<4;i++) {
		  for (j=1;j<5;j++) {
		      fscanf(param_his,"%lg ", &dum);
		      ld->phc->param_his[i][j]=dum;
		      fprintf(stderr, "%g, ",ld->phc->param_his[i][j]);
		  }
	      }
	      fprintf(stderr, "\n");
	      if(param_his!=NULL)
		 fclose(param_his);
	  }
          curr=NULL;
      }	      
      else
	  curr=curr->next;
  }

  if ((temp_ph = fopen("ph.dat", "r")) == NULL)
      fprintf(stderr, "Error opening ph.dat file.\n");

  fscanf(temp_ph, "%lg;\n", &dum);
  ld->ph_value = dum;
  fprintf(stderr, "Read in %g as pH-value\n", ld->ph_value);

  fclose(temp_ph);
  fprintf(stderr, "v1.12: Initializing the ldrec T_ref = %lg m = %lg nu_T = %lg ir->nu_T_lambda %d\n", ld->T_ref, ld->m, ld->nu_T, ir->nu_T_lambda);
}



real fpot(real x, struct t_ldrec *ld){
  /* place a gaussian barrier of 2 kt on the lamda variable */
  real
    a,v,r,rsq,f,b=25;
  r   = (0.5-x);
  rsq = r*r;
  a = 2*ld->T_ref*BOLTZ;
  v   = a*exp(-rsq*b);
  f = -(2*a*b*exp(-b*rsq)*r);
  fprintf(stderr,"pot = %lg, f = %lg\n",v,f);
  return (f);
} /* fpot */



/* currently only for a one dimensional particle */
/* int flag selects the coupling method 
   0 = Andersen thermostat#
   1 = Berendsen
*/
/* return values:
   0 = failure
   1 = success
*/
int tcouple(struct t_ldrec *ld, real* v, t_commrec *cr, int flag, t_inputrec *ir) {
  real rnr=0.0,sigma=0.0, cv=6.90329e-24, lambda=0.0; //cv is currently making tau_t = ld->nu_T
  real kb=1.380658e-23;
  static real tau_t=0.0;
  static gmx_rng_t 
    rng;
  static int start_seed=0;
  
  if(!start_seed){
    /* use the seed initializer to calculate tau for
       berendsen thermostat */
    // divided by one due to 1 particle = 1 deg of freedom
    tau_t = ld->nu_T*1*kb/(2*cv);
    if(MASTER(cr)){
      start_seed = gmx_rng_make_seed();
    }
#ifdef GMX_MPI
    if(PAR(cr)){
      MPI_Bcast(&start_seed,1,MPI_INT,0,MPI_COMM_WORLD);
    }
#endif
		   //rng = gmx_rng_init(42/*start_seed*/); 
		   /* the seed for andersen thermostat is now set in the mdp file. Sdo 17.09.2009*/
		   /* if it is 0 a random number is generated every time */
     if (!ir->andersen_seed)
	rng = gmx_rng_init(start_seed);
     else
        rng = gmx_rng_init((ir->andersen_seed));
  }


		     //fprintf(stderr, "nu_T: %lg, T_ref: %lg, mass: %lg, v: %lg\n", ld->nu_T, ld->T_ref, ld->m, *v);
  /* select da method */
  switch (flag) {
      case 0:
	  /* andersen thermocoupling */
          /* stochastic temperature coupling */
	  //rnr = rnd_gauss_box_mueller();//gmx_rng_uniform_real(rng); 
	  rnr = (real)gmx_rng_uniform_real(rng); /* really uniformely distriubuted
					      number */
	      //fprintf(stderr,"random number generated: %lg and compared to %lg\n",rnr, ld->nu_T);
	  if(fabs(rnr)<ld->nu_T){ 
//fprintf(stderr,"random number really used: %lg\n",rnr);
	      sigma = sqrt(ld->T_ref*BOLTZ/ld->m);
	      *v = gaussdist(rng,sigma);
	  }
	  /**/
	  return 1;
	  break;
      case 1:
	  /* berendsen thermostat */
	  // ld-T contains the temperature of the last step
	  // scale the velocity down by the berendsen lambda factor
	  //fprintf(stderr,"ir->delta_t=%lg tau_t=%lg ld-T_ref=%lg ld->T=%lg BOLTZ=%lg\n", ir->delta_t, tau_t,ld->T_ref,ld->T, BOLTZ);
	  lambda=sqrt(1.0+ir->delta_t/tau_t*(ld->T_ref/ld->T-1.0));
	  //fprintf(stderr,"berendsen-lambda: %lg\n",lambda);
	  if (lambda>1.2) lambda=1.2;
	  if (lambda<0.8) lambda=0.8;
	  *v = *v*lambda;
	  return 1;
	  break;
      default:
	  return 0;
  }
}
real fric(real v){
  return(0);
} /* fric */



/* lambda_4 implies the seperate andersen heat bath,
   theta dynamics and additional dgdl modification
   to mimic correct dgdl behaviour and allow
   constant pH simulations
*/

real update_lambda_4(t_commrec *cr, double dvdl, 
		     t_inputrec *ir, struct t_ldrec *ld,real l_his_partner){
     
  
  int iPos=0;
  real comp;
  static long steps=0; /* very stupid counter */
  struct ph_adapt* tmp_adapt;
  real diff=0.0, ts_now=0.0;
  real dvdl2, radius=0.55, ft=0.0, vn=0.0; /* ft = force on thetha */
  real lambda=0.0;
  double lambda_double=0.0;
  struct lambda_ts* transitions;
  //real beta = 0.0; // from brooks paper constant-ph 1.25 molecular dynamics, Lee et. al
  real corr=0.0;
  real barrier;
  //MPI_Barrier(MPI_COMM_WORLD);
  int i,j;
  i=0;j=0;

  barrier = ld->phc->barrier*(-8); /* barrier for barrier potential in kJ/mol */
  
  /* we set lambba to 0 on all nodes, then compute the actual
     dvdl, and send this to the master (i.e. we sum up all local
     dvdl's ). The master does the integration and obtains the new
     lambda. Via a gmx_sumd, the lambda values gets communicated to
     all nodes. SInce these had lambda=0, they trhus obtain the same
     value as the one on the master.
   */

//  fprintf(stderr, "dvdl before broadcast: %lg\n", dvdl);
  if(PAR(cr)){
    ;//gmx_sumd(1,&dvdl,cr); /* broadcasting the dvdl */ - 26-5-2011 done in md.c
  } 
//  fprintf(stderr, "dvdl after broadcast: %lg\n", dvdl);
  if(MASTER(cr)) {
      fprintf(stderr, "%g, %g, ", dvdl, ld->x); 

      transitions = ld->transitions;

      /* set the transitions value to zero */
      /* the beginning entry is the starting entry from the ld_rec */
      if (steps==0) {
	  /* first step */
	  //fprintf(stderr, "Setting ts entries to zero %i\n", ts_array);
	  transitions[0].lambda = ld->x0;
	  transitions[0].ts = 0.0;
	  for (iPos=1; iPos<=ts_array-1; iPos++)
	  {
	      //fprintf(stderr, "Setting ts on %i\n", iPos);
	      transitions[iPos].ts = 0.0;
	      transitions[iPos].lambda = 0.0;
	  }
      }


    /* correct for the amino_acid's ph correction function */
    //find the corresponding ph function
      tmp_adapt = ld->phc;
      //fprintf(stderr, "%s\n",tmp_adapt->residue); 
      //fprintf(stderr, "Adaption %lf, %lf, %lf, %lf, %lf\n\n", ld->ph_value, tmp_adapt->const_a, tmp_adapt->ph_a, tmp_adapt->const_b, tmp_adapt->ph_b);



      /* first one 
      corr = tmp_adapt->const_a + (tmp_adapt->ph_a * ld->ph_value) + (tmp_adapt->const_b + (tmp_adapt->ph_b * ld->ph_value)) * ld->x;
      */

      /* 20080428 - new params with lambda independent pH correction */

      if (l_his_partner!=-1) {
	  //corr = -(-38.646440685 + 26.819102719 * l_his_partner + (-0.104390885+0.106434888* l_his_partner)* ld->x);//flaten the landscape of his in vacuum
	  //flaten the landscape of his in water (cubic regression) 
	  for (i=0;i<4;i++) {
	      tmp_adapt->param_his[i][0]=tmp_adapt->param_his[i][1]+tmp_adapt->param_his[i][2]*l_his_partner+tmp_adapt->param_his[i][3]*l_his_partner*l_his_partner+tmp_adapt->param_his[i][4]*l_his_partner*l_his_partner*l_his_partner;
	  }
	  i=0;
	  corr = -(tmp_adapt->param_his[i][0]+tmp_adapt->param_his[i+1][0]*ld->x+tmp_adapt->param_his[i+2][0]*ld->x*ld->x+tmp_adapt->param_his[i+3][0]*ld->x*ld->x*ld->x);
          //fprintf(stderr, "corr_flat= %g, ", (corr));
	  if (l_his_partner<0)
	    l_his_partner=0;
	  if (l_his_partner>1)
	    l_his_partner=1;
	  if ( (strncmp( (tmp_adapt->residue),"HIS",3)==0) ) /* four states */
	    corr +=  (1-l_his_partner)*(tmp_adapt->const_b + tmp_adapt->ph_b * ld->ph_value) + l_his_partner*(- tmp_adapt->ph_b*tmp_adapt->second_pka +tmp_adapt->ph_b * ld->ph_value); /* lambda his partner dependent pH correction */  	    
	  if ( (strncmp( (tmp_adapt->residue),"GL2B",4)==0) || (strncmp((tmp_adapt->residue),"AS2B",4)==0) || (strncmp((tmp_adapt->residue),"HI2B",4)==0) ) /* 3 states model */
	    corr+= ( tmp_adapt->const_b - ld->his_partner_const_b ); /* lambda is here the switching coordinate, no pH correction */
	  if ( (strncmp( (tmp_adapt->residue),"GL2A",4)==0) || (strncmp((tmp_adapt->residue),"AS2A",4)==0) || (strncmp((tmp_adapt->residue),"HI2A",4)==0) ) /* 3 states model */ 
	    corr +=  (- tmp_adapt->ph_b *ld->macro_pka + tmp_adapt->ph_b * ld->ph_value); /* lambda his partner dependent pH correction */  
          //fprintf(stderr, "residue=%s;l2=%g;corr_flat+pka= %g,  ",tmp_adapt->residue,l_his_partner, corr);
	  corr += -0.5 * barrier + barrier * ld->x; /* barrier potential */
          //fprintf(stderr, "corr_flat+pka+barrier= %g,  ", corr);
	  //corr += (1-ld->x)*(-1.72)+ld->x*(4.845); /* free energy difference to correct calibration his */
	  //corr += 3.125; /* derivative of correction his */

      }
      else { 
	  
	  corr = -(tmp_adapt->const_a + tmp_adapt->ph_a * ld->x + tmp_adapt->const_2 * (powf(ld->x,2)) + tmp_adapt->const_3 * (powf(ld->x,3)));//flaten the landscape
          corr += -0.5 * barrier + tmp_adapt->const_b + tmp_adapt->ph_b * ld->ph_value + barrier * ld->x;
      }    
	
    /* final dvdl calculation */
    dvdl2 = dvdl + corr;

    ft = dvdl2 * radius*(sin(ld->thetha));

    vn  = ld->v_thetha;
    vn += ft/ld->m*(ir->delta_t); /* vn eq. speed of theta particle */


    //fprintf(stderr, "dvdl2= %g, ", dvdl2);

    //fprintf(stderr, "%g, %g,  ", dvdl2, ft);
    
    /* we want to couple the speed of the lambda-particle to the heat bath */
    /* lambda = r*cos(theta(t)+0.5 */
    /* v_lambda = -r*sin(theta)*vn */
       /*ld->v = -radius*sin(ld->thetha)*vn; */

       /* thermocoupling */
       tcouple(ld, &vn, cr, 0, ir); //andersen
       //tcouple(ld,&vn,cr,1,ir); //berendsen
       
       /* calculate back from the lambda velocity to theta velocity */
       //vn=-(ld->v)/(radius*sin(ld->thetha));

    ld->thetha += vn*(ir->delta_t);
    /* ekin of theta */
    /* ld->ekin = 0.5*ld->m*vn*vn*radius*radius; */
    /* ekin of lambda */
    ld->v = -radius*sin(ld->thetha)*vn; //translation from the v_theta to v_lambda
    ld->ekin = 0.5*ld->m*ld->v*ld->v;
    ld->x = radius*cos(ld->thetha)+0.5;

    ld->T = 2.0*(ld->ekin)/BOLTZ; /* instantaneaous temperature */      
    ld->v_thetha = vn;

    //fprintf(stderr, "%g, %g, ", ld->thetha, ld->v_thetha);

    lambda=(real)ld->x;
    lambda_double=(double)lambda;

    steps++;
    ts_now = (real)steps*(ir->delta_t)/(real)group_count; 
                                              /* steps is static, that means the 
                                                 function is called group_count 
                                                 times in one step */


    /* check for the necessity to adapt the additonal barrier potential */
    //fprintf(stderr, "Test: %lg, %g, %lg, %lg",ts_now, fabs(lambda-transitions[0].lambda), transitions[ts_array-1].ts, barrier);
    /*
    if (transitions[0].lambda <= 0.1)
	comp = 0.0;
    else
	comp = 1.0;
    if ((diff=fabs(lambda-comp)) < 0.1) {
        // this is defined as to be no transition
	// check for the treshold
	if (ts_now-transitions[ts_array-1].ts > adaption_treshold * ts_array) {
	    // adaption of the potential
	    if (barrier >= -5.0)
		barrier = 0.0; //more can't be done
	    else
		barrier = barrier/1.2; //decrease the barrier
	    //resetting the "transition counter" by faking a tranistion in the last box
	    transitions[ts_array-1].ts = ts_now;
	    fprintf(stderr, "No transition occured till ts %g, Barrier is now %g", ts_now, barrier);
	}
    }
    // the intermediate states are not used in any way for the adaptation
    if (diff > 0.9) {
	// transistion occured -> shifting the array, no for loop
	for (iPos=ts_array-1; iPos>0; iPos--) {
	    transitions[iPos].ts = transitions[iPos-1].ts;
	    transitions[iPos].lambda = transitions[iPos-1].lambda;
	}
	transitions[0].ts = ts_now;
	transitions[0].lambda = lambda;
	fprintf(stderr, "Transition from %lg to %lg occured at ts %lg ps",transitions[1].lambda, transitions[0].lambda, ts_now);

	// check for increasing the barrier potential again
        // currently the barrier is reduced when there when the ts_array's transition time is faster than half of the treshold
        // that means: during half of the adaption_treshold time ts_array transitions have occured
	if (ts_now-transitions[ts_array-1].ts < adaption_treshold * ts_array) {
	    if (barrier == 0.0)
		barrier = -5.0;
	    else
		barrier*=1.05; // increase barrier
	    fprintf(stderr, "Additional barrier potential increased to factor: %lg", barrier);
	}
	else
	    fprintf(stderr, "No change in barrier potential required because average transitions are above treshold");
    }

    // prevent that it gets stuck in the middle
    if (diff > 0.1 && diff < 0.9) {
	// in the middle
        // the transisiton counter was last reset in the last lower the barrier operation
	if (ts_now - transitions[ts_array-1].ts > adaption_treshold)
        { 
	    // increase the barrier again
	    
	    if (barrier == 0.0)
		barrier = -5.0;
	    else
		barrier*=1.01; 
	    transitions[ts_array-1].ts = ts_now;
            // increase barrier
	}
    }
    */
    /* fprintf(stderr,"lambda = %lf, total dvdl = %lf + %lf, T_lambda(thetha) = %lg, ekin_lambda = %lg, thetha = %lg, v_thetha = %lg, v_lambda = %lg, ts = %lg\n",lambda,dvdl,dvdl2, ld->T,ld->ekin, ld->thetha, ld->v_thetha, ld->v, ts_now);
     */
  }
  if(PAR(cr)){
    /* communication of lambda value */
    //gmx_sumd(1,&lambda,cr);
    gmx_sumd(1,&lambda_double,cr);
    //fprintf(stderr, "node %d: %g ",cr->nodeid, lambda);
  } 
  lambda=(real)lambda_double;
  
  return ((real)lambda);  
}





real do_lambda_dynamics(t_commrec *cr, double dvdl, 
			t_inputrec *ir,  struct t_ldrec *ld, float l_his_partner ){
  real lambda=0;

  /* only version 4 is supported from 20080221 on */

  /* experimentally fit dgdl to the lambda dynamics function */
      lambda = update_lambda_4(cr,dvdl,ir,ld,l_his_partner);
    /* some printing */
  return lambda;
} /* do_lambda_dynamics */

real lambda_buffer_ions (int charge) {
  int ret;

  return (ret);

}
